@@ -24,8 +24,8 @@
 package com.intel.gkl.compression;
 
 import com.intel.gkl.NativeLibraryLoader;
-import org.apache.logging.log4j.LogManager;
-import org.apache.logging.log4j.Logger;
+import org.apache.commons.logging.LogFactory;
+import org.apache.commons.logging.Log;
 import org.broadinstitute.gatk.nativebindings.NativeLibrary;
 
 import java.io.File;
@@ -35,9 +35,10 @@ import java.util.zip.Deflater;
  * Provides a native Deflater implementation accelerated for the Intel Architecture.
  */
 public final class IntelDeflater extends Deflater implements NativeLibrary {
-    private final static Logger logger = LogManager.getLogger(IntelInflater.class);
+    private final static Log logger = LogFactory.getLog(IntelInflater.class);
     private static final String NATIVE_LIBRARY_NAME = "gkl_compression";
     private static boolean initialized = false;
+    private static final Object lock_class = new Object();
 
     /**
      * Loads the native library, if it is supported on this platform. <p>
@@ -50,15 +51,16 @@ public final class IntelDeflater extends Deflater implements NativeLibrary {
      */
     @Override
     public synchronized boolean load(File tempDir) {
-
-        if (!NativeLibraryLoader.load(tempDir, NATIVE_LIBRARY_NAME)) {
-            return false;
-        }
-        if (!initialized) {
-            initNative();
-            initialized = true;
-        }
-        return true;
+        synchronized (lock_class) {
+		if (!NativeLibraryLoader.load(tempDir, NATIVE_LIBRARY_NAME)) {
+			return false;
+		}
+		if (!initialized) {
+			initNative();
+			initialized = true;
+		}
+	}
+	return true;
     }
 
     private native static void initNative();
@@ -122,10 +124,10 @@ public final class IntelDeflater extends Deflater implements NativeLibrary {
         try {
             resetNative(nowrap);
         } catch (OutOfMemoryError e) {
-            logger.warn("Exception thrown from native Deflater resetNative function call %s", e.getMessage());
+            logger.warn("Exception thrown from native Deflater resetNative function call " + e.getMessage());
             throw new OutOfMemoryError("Memory allocation failed");
         } catch (RuntimeException e) {
-            logger.warn("Exception thrown from native Deflater resetNative function call %s", e.getMessage());
+            logger.warn("Exception thrown from native Deflater resetNative function call " + e.getMessage());
             throw new RuntimeException("Deflate Initialization issue");
         }
 
