@@ -23,8 +23,8 @@
  */
 package com.intel.gkl.smithwaterman;
 
-import org.apache.logging.log4j.LogManager;
-import org.apache.logging.log4j.Logger;
+import org.apache.commons.logging.LogFactory;
+import org.apache.commons.logging.Log;
 import com.intel.gkl.IntelGKLUtils;
 import com.intel.gkl.NativeLibraryLoader;
 
@@ -35,6 +35,7 @@ import org.broadinstitute.gatk.nativebindings.smithwaterman.SWAlignerNativeBindi
 import org.broadinstitute.gatk.nativebindings.smithwaterman.SWNativeAlignerResult;
 
 import java.io.File;
+import java.nio.charset.StandardCharsets;
 
 /**
  * Provides a native SmithWaterman implementation accelerated for the Intel Architecture.
@@ -42,10 +43,11 @@ import java.io.File;
 
 public class IntelSmithWaterman implements SWAlignerNativeBinding {
 
-    private final static Logger logger = LogManager.getLogger(IntelSmithWaterman.class);
+    private final static Log logger = LogFactory.getLog(IntelSmithWaterman.class);
     private static final String NATIVE_LIBRARY_NAME = "gkl_smithwaterman";
     private String nativeLibraryName = "gkl_smithwaterman";
     private static boolean initialized = false;
+    private static final Object lock_class = new Object();
     private IntelGKLUtils gklUtils = new IntelGKLUtils();
 
     // limited due to the internal implementation of the native code in C
@@ -89,10 +91,12 @@ public class IntelSmithWaterman implements SWAlignerNativeBinding {
         if (!NativeLibraryLoader.load(tempDir, NATIVE_LIBRARY_NAME)) {
             return false;
         }
-        if (!initialized) {
 
-            initialized = true;
-        }
+	synchronized (lock_class) {
+		if (!initialized) {
+			initialized = true;
+		}
+	}
 
         if(gklUtils.isAvx512Supported()) {
             logger.info("Using CPU-supported AVX-512 instructions");
@@ -147,14 +151,14 @@ public class IntelSmithWaterman implements SWAlignerNativeBinding {
         try {
             offset = alignNative(refArray, altArray, cigar, parameters.getMatchValue(), parameters.getMismatchPenalty(), parameters.getGapOpenPenalty(), parameters.getGapExtendPenalty(), intStrategy);
         }  catch (OutOfMemoryError e) {
-            logger.warn("Exception thrown from native SW alignNative function call %s", e.getMessage());
+            logger.warn("Exception thrown from native SW alignNative function call " + e.getMessage());
             throw new OutOfMemoryError("Memory allocation failed");
         } catch (IllegalArgumentException e) {
-            logger.warn("Exception thrown from native SW alignNative function call %s", e.getMessage());
+            logger.warn("Exception thrown from native SW alignNative function call " + e.getMessage());
             throw new IllegalArgumentException("Ran into invalid argument issue");
         }
 
-        return new SWNativeAlignerResult(new String(cigar).trim(), offset);
+        return new SWNativeAlignerResult(new String(cigar,StandardCharsets.UTF_8).trim(), offset);
     }
 
     public byte getStrategy(SWOverhangStrategy strategy)
