@@ -11,13 +11,17 @@ import org.testng.annotations.Test;
 
 import java.io.*;
 import java.util.Arrays;
+import java.nio.file.Files;
+import java.nio.charset.StandardCharsets;
+import java.nio.file.Path;
+import java.nio.file.Paths;
 
-import org.apache.logging.log4j.LogManager;
-import org.apache.logging.log4j.Logger;
+import org.apache.commons.logging.LogFactory;
+import org.apache.commons.logging.Log;
 
 public class SmithWatermanUnitTest {
 
-    private final static Logger logger = LogManager.getLogger(SmithWatermanUnitTest.class);
+    private final static Log logger = LogFactory.getLog(SmithWatermanUnitTest.class);
 
     static final String smithwatermanData = IntelGKLUtils.pathToTestResource("smith-waterman.SOFTCLIP.in");
 
@@ -30,45 +34,46 @@ public class SmithWatermanUnitTest {
         if(!isLoaded) {
             String err = "Could not load IntelSmithWaterman; skipping test...";
             logger.warn(err);
+	    smithWaterman.close();
             throw new SkipException(err);
         }
 
         try {
+            final Path Data = Paths.get(smithwatermanData);
+            final BufferedReader in  = new BufferedReader(Files.newBufferedReader(Data, StandardCharsets.UTF_8));
 
-            final File inputFile = new File(smithwatermanData);
-            final FileReader input = new FileReader(inputFile);
-            final BufferedReader in = new BufferedReader(input);
-
-            String refString, altString;
+            String refString = new String(""), altString = new String("");
             SWParameters SWparameters = new SWParameters(200, -150, -260, -11);
             //SWParameters SWparameters = new SWParameters(3, -1, -4, -3);
             SWOverhangStrategy SWstrategy = SWOverhangStrategy.SOFTCLIP;
-
-            refString = in.readLine();
-            altString = in.readLine();
-
+	    try {
+            	refString = in.readLine();
+            	altString = in.readLine();
+	    } catch (IOException e) {System.err.println("Caught IOException: " +  e.getMessage());}
             try {
-                smithWaterman.align(null, altString.getBytes(), SWparameters, SWstrategy);
+                if (altString != null)
+			smithWaterman.align(null, altString.getBytes("UTF-8"), SWparameters, SWstrategy);
                 Assert.fail("NullPointerException expected.");
             } catch(NullPointerException e) {}
 
             try {
-                smithWaterman.align(refString.getBytes(), null, SWparameters, SWstrategy);
+                smithWaterman.align(refString.getBytes("UTF-8"), null, SWparameters, SWstrategy);
                 Assert.fail("NullPointerException expected.");
             } catch(NullPointerException e) {}
 
             try {
-                smithWaterman.align(refString.getBytes(), altString.getBytes(), null, SWstrategy);
+                smithWaterman.align(refString.getBytes("UTF-8"), altString.getBytes("UTF-8"), null, SWstrategy);
                 Assert.fail("NullPointerException expected.");
             } catch(NullPointerException e) {}
 
             try {
-                smithWaterman.align(refString.getBytes(), altString.getBytes(), SWparameters, null);
+                smithWaterman.align(refString.getBytes("UTF-8"), altString.getBytes("UTF-8"), SWparameters, null);
                 Assert.fail("NullPointerException expected.");
             } catch(NullPointerException e) {}
-        } catch (java.io.IOException e) {
+	} catch (java.io.IOException e) {
             e.printStackTrace();
-        }
+	}
+	smithWaterman.close();
     }
 
     @Test(expectedExceptions = IllegalArgumentException.class)
@@ -142,28 +147,33 @@ public class SmithWatermanUnitTest {
     public void singleElementSequencesAlignmentTest(){
         final IntelSmithWaterman sw = new IntelSmithWaterman();
         boolean isLoaded = sw.load(null);
-        if(!isLoaded) throw new SkipException("Could not load IntelSmithWaterman; skipping test...");
-
+        if(!isLoaded) {
+		sw.close();
+		throw new SkipException("Could not load IntelSmithWaterman; skipping test...");
+	}
         byte[] ref =    new byte [] {'C'};
         byte[] align =  new byte [] {'C'};
 
         SWParameters SWparameters = new SWParameters( 3,-2,-2,-1);
         SWNativeAlignerResult result = sw.align(ref, align, SWparameters, SWOverhangStrategy.IGNORE);
-
+	sw.close();
         Assert.assertEquals(result.cigar, "1M");
     }
     @Test(enabled = true)
     public void twoElementSequencesAlignmentTest(){
         final IntelSmithWaterman sw = new IntelSmithWaterman();
         boolean isLoaded = sw.load(null);
-        if(!isLoaded) throw new SkipException("Could not load IntelSmithWaterman; skipping test...");
-
+        if(!isLoaded) {
+		sw.close();
+		throw new SkipException("Could not load IntelSmithWaterman; skipping test...");
+	}
         byte[] ref =    new byte [] {'A', 'D'};
         byte[] align =  new byte [] {'A', 'T'};
 
         SWParameters SWparameters = new SWParameters( 3,-5,-2,-1);
         SWNativeAlignerResult result = sw.align(ref, align, SWparameters, SWOverhangStrategy.IGNORE);
 
+	sw.close();
         Assert.assertEquals(result.cigar, "1M1I");
     }
 
@@ -172,8 +182,10 @@ public class SmithWatermanUnitTest {
         final IntelSmithWaterman smithWaterman = new IntelSmithWaterman();
         final boolean isLoaded = smithWaterman.load(null);
         
-        if(!isLoaded) throw new SkipException("Could not load IntelSmithWaterman; skipping test...");
-
+        if(!isLoaded) {
+		smithWaterman.close();
+		throw new SkipException("Could not load IntelSmithWaterman; skipping test...");
+	}
         int matchValue = TestingUtils.MAX_SW_MATCH_VALUE;
         int sequenceLength = TestingUtils.MAX_SW_SEQUENCE_LENGTH;
 
@@ -185,7 +197,7 @@ public class SmithWatermanUnitTest {
         SWParameters SWparameters = new SWParameters(matchValue, -5, -10, -10);
         SWNativeAlignerResult result = smithWaterman.align(ref, align, SWparameters, SWOverhangStrategy.IGNORE);
 
-
+	smithWaterman.close();
         Assert.assertEquals(result.cigar, expectedCigar);
     }
 
@@ -200,42 +212,47 @@ public class SmithWatermanUnitTest {
         if(!isLoaded) {
             String err = "Could not load IntelSmithWaterman; skipping test...";
             logger.warn(err);
-            throw new SkipException(err);
+            smithWaterman.close();
+	    throw new SkipException(err);
         }
-
         try {
-
-            final File inputFile = new File(smithwatermanData);
-            long inputBytes = inputFile.length();
-            final FileReader input = new FileReader(inputFile);
-            final BufferedReader in = new BufferedReader(input);
+            final Path Data = Paths.get(smithwatermanData);
+            final BufferedReader in  = new BufferedReader(Files.newBufferedReader(Data, StandardCharsets.UTF_8));
 
             byte[] ref;
             byte[] alt;
 
-            String refString, altString;
+            String refString = new String(""), altString = new String("");
             SWParameters SWparameters = new SWParameters(200, -150, -260, -11);
             //SWParameters SWparameters = new SWParameters(3, -1, -4, -3);
             SWOverhangStrategy SWstrategy = SWOverhangStrategy.SOFTCLIP;
 
+	    try {
+		 refString = in.readLine();
+	    } catch (IOException e) {System.err.println("Caught IOException: " +  e.getMessage());}
 
-            while(in.readLine() !=null) {
+	   while(refString !=null) {
 
-               refString = in.readLine();
-                ref = refString.getBytes();
+		ref = refString.getBytes("UTF-8");
+		try {
+			altString = in.readLine();
+		} catch (IOException e) {System.err.println("Caught IOException: " +  e.getMessage());}
 
-                altString = in.readLine();
-               alt = altString.getBytes();
-
-
-                //Get the results for one pair
-                SWNativeAlignerResult result = smithWaterman.align(refString.getBytes(), altString.getBytes(), SWparameters, SWstrategy);
+		if (altString !=null) {
+			alt = altString.getBytes("UTF-8");
+			//Get the results for one pair
+			SWNativeAlignerResult result = smithWaterman.align(refString.getBytes("UTF-8"), altString.getBytes("UTF-8"), SWparameters, SWstrategy);
+		}
+		try {
+			refString = in.readLine();
+		} catch (IOException e) {System.err.println("Caught IOException: " +  e.getMessage());}
 
             }
-
+	    in.close();
         } catch (java.io.IOException e) {
-            e.printStackTrace();
+		e.printStackTrace();
         }
+	smithWaterman.close();
     }
 }
 
